#!/usr/bin/env ruby
##
# Copyright (c) 2013 Christian Ferraz Lemos de Sousa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
# INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
# PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
# CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
# OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
####
# Base Monk classes
##

# monk needs mongo (duh)
require 'mongo'

module Sinatra
  module Monk
    # Provides a base class for Mongo database management. This is the simplest way
    # to connect to Mongo with Monk. It is intented to keep Sinatra's philosofy of
    # keeping it as simple as possible for use.
    class MBase
      attr_reader :client, :database, :username, :password

      # (Any Ruby) Setup the Monk with the data that was sent. Autoconnect if :connect flag is true.
      # @param user [String] the database username,
      # @param pass [String] the password of the previous database user,
      # @param database [String] the name of the database,
      # @param host [String] the host of the database,
      # @param port [Fixnum] the port to connect to the database,
      # @param opts [Hash] monk options (set :conect to autoconnect the Monk)
      def initialize user = '', pass = '', database = 'local', host = 'localhost',
                     port = 27017, opts = {:connect => false}
        # these data can't just be changed runtime
        @client        = Mongo::MongoClient.new 'localhost', 27017
        @database      = @client[database]
        @database_name = database

        # non persistent data (can be changed before connect)
        @username      = user
        @password      = pass

        # connect if asked
        connect if opts[:connect]
      end

      # Closes the connection with the server, you MAY NOT TRY to alter the collection while
      # in that state.
      def close
        @client.close
      end

      # Returns the Monk connection state.
      def connected?
        return @client.connected?
      end

      # Conects the Monk to Mongo.
      # @param user [String] the database username,
      # @param pass [String] the password of the previous database user.
      def connect user = @username, pass = @password
        return true if connected?
        @client.connect
        @username  = user
        @password  = pass
        return @database.authenticate @username, @password
      end
    end

    # A type of Monk that acts like a Hash, it was made to look very native.
    # It still need some improvement.
    class MHash < MBase
      attr_reader :local, :name, :collection
      attr_accessor :sync

      # (Any Ruby) Differently from MBase, MHash enforce autoconnect.
      # @param name [String] name of the used/created collection,
      # @param user [String] the database username,
      # @param pass [String] the password of the previous database user,
      # @param database [String] the name of the database,
      # @param host [String] the host of the database,
      # @param port [Fixnum] the port to connect to the database,
      # @param opts [Hash] monk options (set :conect to autoconnect the Monk)
      def initialize name = 'monk', user = '', pass = '', database = 'local',
        host = 'localhost', port = 27017, opts = {:sync => true}
        # forced premature connection (option override)
        opts[:connect]= true

        # HashMonk special data
        @name         = name         # collection name
        @sync         = opts[:sync]  # autosave collection
        @local        = Hash.new     # local information

        # need update flag (writes local data to db)
        @need_sync    = false

        # build our monk
        super host, port, user, pass, opts

        install if connect
      end

      # Sync the local collection with the remote one.
      def sync
        if @need_sync
          @database[@name].save(@collection)
          @need_sync = false
          return true
        end
        return false
      end

      # Returns true if collection need synchronization.
      def need_sync?
        return @need_sync
      end

      # Returns true if collection has a key.
      # @param key [String] the key.
      def include?(key)
        return @collection['content'].has_key? key
      end

      # Returns the collection inspect data.
      def inspect
        return @collection['content'].inspect
      end

      # Delete a key from the collection.
      # @param key [String] the key.
      def delete(key)
        raise TypeError, "#{key} can't be coerced into String" unless key.is_a? String
        if @collection['content'].has_key? key
          @collection['content'].delete key
          need_sync!
          return true
        end
        return false
      end

      # Gives read access to a key in the collection.
      # @param key [String] the key.
      def [](key)
        raise TypeError, "#{key} can't be coerced into String" unless key.is_a? String
        return @collection['content'][key] if @collection['content'].has_key? key
        raise InexistentKey, "Field #{key} not found."
      end

      # Gives write access to a key in the collection.
      # @param key [String] the key.
      def []=(key, value)
        raise TypeError, "#{key} can't be coerced into String" unless key.is_a? String
        @collection['content'][key] = value
        need_sync!
      end

      private
      # Install the collection in the database (may not install if already exists,
      # in that case it just uses it).
      def install
        if @database.collection_names.include? @name
          # setup for use existent collection
          @collection = @database[@name].find_one({'name'=>@name})
        else
          # install collection
          @database.create_collection @name
          @database[@name].insert([{'name' => @name,
                                           'content' => {}}])
          @collection = @database[@name].find_one({'name'=>@name})
          need_sync!
        end
      end

      # Raises a sync flag, if autosync mode is on, it just sync the database.
      def need_sync!
        @need_sync = true
        sync if @sync # autosync
      end
    end
  end
end
